@@ -1,18 +1,20 @@
-import re
-import ssl
-import enum
-import socket
+# Copyright 2014-2021 The aiosmtpd Developers
+# SPDX-License-Identifier: Apache-2.0
+
 import asyncio
-import inspect
-import logging
+import asyncio.sslproto as sslproto
 import binascii
 import collections
-import asyncio.sslproto as sslproto
-
+import enum
+import inspect
+import logging
+import re
+import socket
+import ssl
+from aiosmtpd import __version__
 from base64 import b64decode, b64encode
 from email._header_value_parser import get_addr_spec, get_angle_addr
 from email.errors import HeaderParseError
-from public import public
 from typing import (
     Any,
     AnyStr,
@@ -23,10 +25,14 @@ from typing import (
     List,
     NamedTuple,
     Optional,
+    Tuple,
     Union,
 )
 from warnings import warn
 
+import attr
+from public import public
+
 
 # region #### Custom Data Types #######################################################
 
@@ -46,7 +52,8 @@ class _DataState(enum.Enum):
     TOO_MUCH = enum.auto()
 
 
-AuthCallbackType = Callable[[str, Optional[bytes], Optional[bytes]], bool]
+AuthCallbackType = Callable[[str, bytes, bytes], bool]
+AuthenticatorType = Callable[["SMTP", "Session", "Envelope", str, Any], "AuthResult"]
 AuthMechanismType = Callable[["SMTP", List[str]], Awaitable[Any]]
 _TriStateType = Union[None, _Missing, bytes]
 
@@ -60,8 +67,8 @@ __all__ = [
     "AuthCallbackType",
     "AuthMechanismType",
     "MISSING",
+    "__version__",
 ]  # Will be added to by @public
-__version__ = '1.2.4'
 __ident__ = 'Python SMTP {}'.format(__version__)
 log = logging.getLogger('mail.log')
 
@@ -90,10 +97,45 @@ CLIENT_AUTH_B = re.compile(
     # stripped line. If no bCRLF, then this group will be b""
     br"(?P<crlf>(?:\r\n)?)", re.IGNORECASE
 )
+"""Regex that matches 'AUTH <mech> <param>' commend"""
 
 # endregion
 
 
+@attr.s
+class AuthResult:
+    """
+    Contains the result of authentication, to be returned to the smtp_AUTH method.
+    All initialization arguments _must_ be keyworded!
+    """
+
+    success: bool = attr.ib(kw_only=True)
+    """Indicates authentication is successful or not"""
+
+    handled: bool = attr.ib(kw_only=True, default=True)
+    """
+    True means everything (including sending of status code) has been handled by the
+    AUTH handler and smtp_AUTH should not do anything else.
+    Applicable only if success == False.
+    """
+
+    message: Optional[str] = attr.ib(kw_only=True, default=None)
+    """Optional message for additional handling by smtp_AUTH"""
+
+    auth_data: Optional[Any] = attr.ib(kw_only=True, default=None)
+    """
+    Optional free-form authentication data. For the built-in mechanisms, it is usually
+    an instance of LoginPassword. Other implementations are free to use any data
+    structure here.
+    """
+
+
+@public
+class LoginPassword(NamedTuple):
+    login: bytes
+    password: bytes
+
+
 @public
 class Session:
     def __init__(self, loop):
@@ -102,11 +144,19 @@ class Session:
         self.host_name = None
         self.extended_smtp = False
         self.loop = loop
-        self.login_data = None
 
-    @property
-    def authenticated(self):
-        return self.login_data is not None
+        self.login_data = None
+        """Legacy login_data, usually containing the username"""
+
+        self.auth_data = None
+        """
+        New system *optional* authentication data;
+        can contain anything returned by the authenticator callback.
+        Can even be None; check `authenticated` attribute to determine
+        if AUTH successful or not.
+        """
+
+        self.authenticated = None
 
 
 @public
@@ -131,7 +181,8 @@ def make_loop():
 @public
 def syntax(text, extended=None, when: Optional[str] = None):
     """
-    Provides helptext for (E)SMTP HELP. Applies for smtp_* methods only!
+    A @decorator that provides helptext for (E)SMTP HELP.
+    Applies for smtp_* methods only!
 
     :param text: Help text for (E)SMTP HELP
     :param extended: Additional text for ESMTP HELP (appended to text)
@@ -149,7 +200,7 @@ def syntax(text, extended=None, when: Optional[str] = None):
 @public
 def auth_mechanism(actual_name: str):
     """
-    Explicitly specifies the name of the AUTH mechanism implemented by
+    A @decorator to explicitly specifies the name of the AUTH mechanism implemented by
     the function/method this decorates
 
     :param actual_name: Name of AUTH mechanism. Must consists of [A-Z0-9_-] only.
@@ -164,7 +215,9 @@ def auth_mechanism(actual_name: str):
     return decorator
 
 
-def login_always_fail(mechanism, login, password):
+def login_always_fail(
+        mechanism: str, login: bytes, password: bytes
+) -> bool:
     return False
 
 
@@ -187,6 +240,9 @@ def sanitize(text: bytes) -> bytes:
 
 @public
 def sanitized_log(func: Callable, msg: AnyStr, *args, **kwargs):
+    """
+    Sanitize args before passing to a logging function.
+    """
     sanitized_args = [
         sanitize(a) if isinstance(a, bytes) else a
         for a in args
@@ -197,7 +253,7 @@ def sanitized_log(func: Callable, msg: AnyStr, *args, **kwargs):
 @public
 class SMTP(asyncio.StreamReaderProtocol):
     command_size_limit = 512
-    command_size_limits = collections.defaultdict(
+    command_size_limits: Dict[str, int] = collections.defaultdict(
         lambda x=command_size_limit: x)
     line_length_limit = 1001
     """Maximum line length according to RFC 5321 s 4.5.3.1.6"""
@@ -205,24 +261,26 @@ class SMTP(asyncio.StreamReaderProtocol):
     # (RFC 5322 s 2.1.1 + RFC 6532 s 3.4) 998 octets + CRLF = 1000 octets
     # (RFC 5321 s 4.5.3.1.6) 1000 octets + "transparent dot" = 1001 octets
 
-    def __init__(
-            self, handler,
-            *,
-            data_size_limit=DATA_SIZE_DEFAULT,
-            enable_SMTPUTF8=False,
-            decode_data=False,
-            hostname=None,
-            ident=None,
-            tls_context: Optional[ssl.SSLContext] = None,
-            require_starttls=False,
-            timeout=300,
-            auth_required=False,
-            auth_require_tls=True,
-            auth_exclude_mechanism: Optional[Iterable[str]] = None,
-            auth_callback: AuthCallbackType = None,
-            command_call_limit: Union[int, Dict[str, int], None] = None,
-            loop=None,
-    ):
+    AuthLoginUsernameChallenge = "User Name\x00"
+    AuthLoginPasswordChallenge = "Password\x00"
+
+    def __init__(self, handler,
+                 *,
+                 data_size_limit=DATA_SIZE_DEFAULT,
+                 enable_SMTPUTF8=False,
+                 decode_data=False,
+                 hostname=None,
+                 ident=None,
+                 tls_context: Optional[ssl.SSLContext] = None,
+                 require_starttls=False,
+                 timeout=300,
+                 auth_required=False,
+                 auth_require_tls=True,
+                 auth_exclude_mechanism: Optional[Iterable[str]] = None,
+                 auth_callback: AuthCallbackType = None,
+                 command_call_limit: Union[int, Dict[str, int], None] = None,
+                 authenticator: AuthenticatorType = None,
+                 loop=None):
         self.__ident__ = ident or __ident__
         self.loop = loop if loop else make_loop()
         super().__init__(
@@ -264,8 +322,18 @@ class SMTP(asyncio.StreamReaderProtocol):
                  "can lead to security vulnerabilities!")
             log.warning("auth_required == True but auth_require_tls == False")
         self._auth_require_tls = auth_require_tls
-        self._auth_callback = auth_callback or login_always_fail
+
+        self._authenticator: Optional[AuthenticatorType]
+        self._auth_callback: Optional[AuthCallbackType]
+        if authenticator is not None:
+            self._authenticator = authenticator
+            self._auth_callback = None
+        else:
+            self._auth_callback = auth_callback or login_always_fail
+            self._authenticator = None
+
         self._auth_required = auth_required
+
         # Get hooks & methods to significantly speedup getattr's
         self._auth_methods: Dict[str, _AuthMechAttr] = {
             getattr(
@@ -290,24 +358,45 @@ class SMTP(asyncio.StreamReaderProtocol):
             for m in dir(handler)
             if m.startswith("handle_")
         }
+
+        # When we've deprecated the 4-arg form of handle_EHLO,
+        # we can -- and should -- remove this whole code block
+        ehlo_hook = self._handle_hooks.get("EHLO", None)
+        if ehlo_hook is None:
+            self._ehlo_hook_ver = None
+        else:
+            ehlo_hook_params = inspect.signature(ehlo_hook).parameters
+            if len(ehlo_hook_params) == 4:
+                self._ehlo_hook_ver = "old"
+                warn("Use the 5-argument handle_EHLO() hook instead of "
+                     "the 4-argument handle_EHLO() hook; "
+                     "support for the 4-argument handle_EHLO() hook will be "
+                     "removed in version 2.0",
+                     DeprecationWarning)
+            elif len(ehlo_hook_params) == 5:
+                self._ehlo_hook_ver = "new"
+            else:
+                raise RuntimeError("Unsupported EHLO Hook")
+
         self._smtp_methods: Dict[str, Any] = {
             m.replace("smtp_", ""): getattr(self, m)
             for m in dir(self)
             if m.startswith("smtp_")
         }
 
+        self._call_limit_default: int
         if command_call_limit is None:
             self._enforce_call_limit = False
         else:
             self._enforce_call_limit = True
             if isinstance(command_call_limit, int):
                 self._call_limit_base = {}
-                self._call_limit_default: int = command_call_limit
+                self._call_limit_default = command_call_limit
             elif isinstance(command_call_limit, dict):
                 if not all(map(is_int, command_call_limit.values())):
                     raise TypeError("All command_call_limit values must be int")
                 self._call_limit_base = command_call_limit
-                self._call_limit_default: int = command_call_limit.get(
+                self._call_limit_default = command_call_limit.get(
                     "*", CALL_LIMIT_DEFAULT
                 )
             else:
@@ -632,32 +721,50 @@ class SMTP(asyncio.StreamReaderProtocol):
         if not hostname:
             await self.push('501 Syntax: EHLO hostname')
             return
+
+        response = []
         self._set_rset_state()
         self.session.extended_smtp = True
-        await self.push('250-%s' % self.hostname)
+        response.append('250-%s' % self.hostname)
         if self.data_size_limit:
-            await self.push('250-SIZE %s' % self.data_size_limit)
+            response.append('250-SIZE %s' % self.data_size_limit)
             self.command_size_limits['MAIL'] += 26
         if not self._decode_data:
-            await self.push('250-8BITMIME')
+            response.append('250-8BITMIME')
         if self.enable_SMTPUTF8:
-            await self.push('250-SMTPUTF8')
+            response.append('250-SMTPUTF8')
             self.command_size_limits['MAIL'] += 10
         if self.tls_context and not self._tls_protocol:
-            await self.push('250-STARTTLS')
+            response.append('250-STARTTLS')
+        if not self._auth_require_tls or self._tls_protocol:
+            response.append(
+                "250-AUTH " + " ".join(sorted(self._auth_methods.keys()))
+            )
+
         if hasattr(self, 'ehlo_hook'):
             warn('Use handler.handle_EHLO() instead of .ehlo_hook()',
                  DeprecationWarning)
             await self.ehlo_hook()
-        if not self._auth_require_tls or self._tls_protocol:
-            await self.push(
-                "250-AUTH " + " ".join(sorted(self._auth_methods.keys()))
-            )
-        status = await self._call_handler_hook('EHLO', hostname)
-        if status is MISSING:
+
+        if self._ehlo_hook_ver is None:
             self.session.host_name = hostname
-            status = '250 HELP'
-        await self.push(status)
+            response.append('250 HELP')
+        elif self._ehlo_hook_ver == "old":
+            # Old behavior: Send all responses first...
+            for r in response:
+                await self.push(r)
+            # ... then send the response from the hook.
+            response = [await self._call_handler_hook("EHLO", hostname)]
+            # (The hook might internally send its own responses.)
+        elif self._ehlo_hook_ver == "new":  # pragma: nobranch
+            # New behavior: hand over list of responses so far to the hook, and
+            # REPLACE existing list of responses with what the hook returns.
+            # We will handle the push()ing
+            response.append('250 HELP')
+            response = await self._call_handler_hook("EHLO", hostname, response)
+
+        for r in response:
+            await self.push(r)
 
     @syntax('NOOP [ignored]')
     async def smtp_NOOP(self, arg):
@@ -715,49 +822,71 @@ class SMTP(asyncio.StreamReaderProtocol):
         if await self.check_helo_needed("EHLO"):
             return
         elif not self.session.extended_smtp:
-            await self.push("500 Error: command 'AUTH' not recognized")
+            return await self.push("500 Error: command 'AUTH' not recognized")
         elif self._auth_require_tls and not self._tls_protocol:
-            await self.push("538 5.7.11 Encryption required for requested "
-                            "authentication mechanism")
+            return await self.push("538 5.7.11 Encryption required for requested "
+                                   "authentication mechanism")
         elif self.session.authenticated:
-            await self.push('503 Already authenticated')
+            return await self.push('503 Already authenticated')
         elif not arg:
-            await self.push('501 Not enough value')
-        else:
-            args = arg.split()
-            if len(args) > 2:
-                await self.push('501 Too many values')
-                return
+            return await self.push('501 Not enough value')
 
-            mechanism = args[0]
-            if mechanism not in self._auth_methods:
-                await self.push('504 5.5.4 Unrecognized authentication type')
-                return
+        args = arg.split()
+        if len(args) > 2:
+            return await self.push('501 Too many values')
 
-            status = await self._call_handler_hook('AUTH', args)
-            if status is MISSING:
-                method = self._auth_methods[mechanism]
-                if method.is_builtin:
-                    log.debug(f"Using builtin auth_ hook for {mechanism}")
+        mechanism = args[0]
+        if mechanism not in self._auth_methods:
+            return await self.push('504 5.5.4 Unrecognized authentication type')
+
+        CODE_SUCCESS = "235 2.7.0 Authentication successful"
+        CODE_INVALID = "535 5.7.8 Authentication credentials invalid"
+        status = await self._call_handler_hook('AUTH', args)
+        if status is MISSING:
+            auth_method = self._auth_methods[mechanism]
+            if auth_method.is_builtin:
+                log.debug(f"Using builtin auth_ hook for {mechanism}")
+            else:
+                log.debug(f"Using handler auth_ hook for {mechanism}")
+            # Pass 'self' to method so external methods can leverage this
+            # class's helper methods such as push()
+            auth_result = await auth_method.method(self, args)
+            log.debug(f"auth_{mechanism} returned {auth_result}")
+
+            # New system using `authenticator` and AuthResult
+            if isinstance(auth_result, AuthResult):
+                if auth_result.success:
+                    self.session.authenticated = True
+                    _auth_data = auth_result.auth_data
+                    self.session.auth_data = _auth_data
+                    # Custom mechanisms might not implement the "login" attribute, and
+                    # that's okay.
+                    self.session.login_data = getattr(_auth_data, "login", None)
+                    status = auth_result.message or CODE_SUCCESS
                 else:
-                    log.debug(f"Using handler auth_ hook for {mechanism}")
-                # Pass 'self' to method so external methods can leverage this
-                # class's helper methods such as push()
-                login_data = await method.method(self, args)
-                log.debug(f"auth_{mechanism} returned {login_data}")
-                if login_data is None:
-                    # None means there's an error already handled by method and
-                    # we don't need to do anything more
-                    return
-                elif login_data is MISSING:
-                    # MISSING means no error in AUTH process, but credentials
-                    # is rejected / not valid
-                    status = '535 5.7.8 Authentication credentials invalid'
-                else:
-                    self.session.login_data = login_data
-                    status = '235 2.7.0 Authentication successful'
-            if status is not None:  # pragma: no branch
-                await self.push(status)
+                    if auth_result.handled:
+                        status = None
+                    elif auth_result.message:
+                        status = auth_result.message
+                    else:
+                        status = CODE_INVALID
+
+            # Old system using `auth_callback` and _TriState
+            elif auth_result is None:
+                # None means there's an error already handled by method and
+                # we don't need to do anything more
+                status = None
+            elif auth_result is MISSING or auth_result is False:
+                # MISSING means no error in AUTH process, but credentials
+                # is rejected / not valid
+                status = CODE_INVALID
+            else:
+
+                self.session.login_data = auth_result
+                status = CODE_SUCCESS
+
+        if status is not None:  # pragma: no branch
+            await self.push(status)
 
     async def challenge_auth(
             self,
@@ -822,36 +951,54 @@ class SMTP(asyncio.StreamReaderProtocol):
             encode_to_b64=False,
         )
 
+    def _authenticate(self, mechanism, auth_data) -> AuthResult:
+        if self._authenticator is not None:
+            # self.envelope is likely still empty, but we'll pass it anyways to
+            # make the invocation similar to the one in _call_handler_hook
+            return self._authenticator(
+                self, self.session, self.envelope, mechanism, auth_data
+            )
+        else:
+            assert self._auth_callback is not None
+            assert isinstance(auth_data, LoginPassword)
+            if self._auth_callback(mechanism, *auth_data):
+                return AuthResult(success=True, handled=True, auth_data=auth_data)
+            else:
+                return AuthResult(success=False, handled=False)
+
     # IMPORTANT NOTES FOR THE auth_* METHODS
+    # ======================================
+    # Please note that there are two systems for return values in #2.
     #
-    # 1. For internal methods, due to how they are called, we must ignore
-    #    the first arg
-    # 2. All auth_* methods can return one of three values:
+    # 1. For internal methods, due to how they are called, we must ignore the first arg
+    # 2. (OLD SYSTEM) All auth_* methods can return one of three values:
     #    - None: An error happened and handled;
     #            smtp_AUTH should do nothing more
-    #    - MISSING: No error during SMTP AUTH process, but authentication
-    #               failed
+    #    - MISSING or False: Authentication failed, but not because of error
     #    - [Any]: Authentication succeeded and this is the 'identity' of
     #             the SMTP user
     #      - 'identity' is not always username, depending on the auth mecha-
     #        nism. Might be a session key, a one-time user ID, or any kind of
     #        object, actually.
+    # 2. (NEW SYSTEM) All auth_* methods must return an AuthResult object.
+    #    For explanation on the object's attributes,
+    #    see the AuthResult class definition.
     # 3. Auth credentials checking is performed in the auth_* methods because
     #    more advanced auth mechanism might not return login+password pair
     #    (see #2 above)
 
-    async def auth_PLAIN(self, _, args: List[str]):
+    async def auth_PLAIN(self, _, args: List[str]) -> AuthResult:
         login_and_password: _TriStateType
         if len(args) == 1:
             login_and_password = await self.challenge_auth("")
             if login_and_password is MISSING:
-                return
+                return AuthResult(success=False)
         else:
             try:
                 login_and_password = b64decode(args[1].encode(), validate=True)
             except Exception:
                 await self.push("501 5.5.2 Can't decode base64")
-                return
+                return AuthResult(success=False, handled=True)
         try:
             # login data is "{authz_id}\x00{login_id}\x00{password}"
             # authz_id can be null, and currently ignored
@@ -859,32 +1006,35 @@ class SMTP(asyncio.StreamReaderProtocol):
             _, login, password = login_and_password.split(b"\x00")
         except ValueError:  # not enough args
             await self.push("501 5.5.2 Can't split auth value")
-            return
+            return AuthResult(success=False, handled=True)
         # Verify login data
         assert login is not None
         assert password is not None
-        if self._auth_callback("PLAIN", login, password):
-            return login
-        else:
-            return MISSING
+        return self._authenticate("PLAIN", LoginPassword(login, password))
 
     async def auth_LOGIN(self, _, args: List[str]):
         login: _TriStateType
-        login = await self.challenge_auth(b"User Name\x00")
-        if login is MISSING:
-            return
+        if len(args) == 1:
+            # Client sent only "AUTH LOGIN"
+            login = await self.challenge_auth(self.AuthLoginUsernameChallenge)
+            if login is MISSING:
+                return AuthResult(success=False)
+        else:
+            # Client sent "AUTH LOGIN <b64-encoded-username>"
+            try:
+                login = b64decode(args[1].encode(), validate=True)
+            except Exception:
+                await self.push("501 5.5.2 Can't decode base64")
+                return AuthResult(success=False, handled=True)
+        assert login is not None
 
         password: _TriStateType
-        password = await self.challenge_auth(b"Password\x00")
+        password = await self.challenge_auth(self.AuthLoginPasswordChallenge)
         if password is MISSING:
-            return
-
-        assert login is not None
+            return AuthResult(success=False)
         assert password is not None
-        if self._auth_callback("LOGIN", login, password):
-            return login
-        else:
-            return MISSING
+
+        return self._authenticate("LOGIN", LoginPassword(login, password))
 
     def _strip_command_keyword(self, keyword, arg):
         keylen = len(keyword)
@@ -892,19 +1042,28 @@ class SMTP(asyncio.StreamReaderProtocol):
             return arg[keylen:].strip()
         return None
 
-    def _getaddr(self, arg):
+    def _getaddr(self, arg) -> Tuple[Optional[str], Optional[str]]:
+        """
+        Try to parse address given in SMTP command.
+
+        Returns address=None if arg can't be parsed properly (get_angle_addr /
+        get_addr_spec raised HeaderParseError)
+        """
         if not arg:
             return '', ''
-        if arg.lstrip().startswith('<'):
-            address, rest = get_angle_addr(arg)
-        else:
-            address, rest = get_addr_spec(arg)
         try:
-            address = address.addr_spec
-        except IndexError:
-            # Workaround http://bugs.python.org/issue27931
-            address = None
-        return address, rest
+            if arg.lstrip().startswith('<'):
+                address, rest = get_angle_addr(arg)
+            else:
+                address, rest = get_addr_spec(arg)
+        except HeaderParseError:
+            return None, None
+        address = address.addr_spec
+        localpart, atsign, domainpart = address.rpartition("@")
+        if len(localpart) > 64:  # RFC 5321 § 4.5.3.1.1
+            return None, None
+        else:
+            return address, rest
 
     def _getparams(self, params):
         # Return params as dictionary. Return None if not all parameters
@@ -952,10 +1111,7 @@ class SMTP(asyncio.StreamReaderProtocol):
         if await self.check_auth_needed("VRFY"):
             return
         if arg:
-            try:
-                address, params = self._getaddr(arg)
-            except HeaderParseError:
-                address = None
+            address, params = self._getaddr(arg)
             if address is None:
                 await self.push('502 Could not VRFY %s' % arg)
             else:
@@ -983,17 +1139,18 @@ class SMTP(asyncio.StreamReaderProtocol):
         if arg is None:
             await self.push(syntaxerr)
             return
-        address, params = self._getaddr(arg)
+        address, addrparams = self._getaddr(arg)
         if address is None:
-            await self.push(syntaxerr)
-            return
-        if not self.session.extended_smtp and params:
+            return await self.push("553 5.1.3 Error: malformed address")
+        if not address:
+            return await self.push(syntaxerr)
+        if not self.session.extended_smtp and addrparams:
             await self.push(syntaxerr)
             return
         if self.envelope.mail_from:
             await self.push('503 Error: nested MAIL command')
             return
-        mail_options = params.upper().split()
+        mail_options = addrparams.upper().split()
         params = self._getparams(mail_options)
         if params is None:
             await self.push(syntaxerr)
@@ -1041,35 +1198,33 @@ class SMTP(asyncio.StreamReaderProtocol):
         if await self.check_auth_needed("RCPT"):
             return
         if not self.envelope.mail_from:
-            await self.push('503 Error: need MAIL command')
-            return
+            return await self.push('503 Error: need MAIL command')
+
         syntaxerr = '501 Syntax: RCPT TO: <address>'
         if self.session.extended_smtp:
             syntaxerr += ' [SP <mail-parameters>]'
         if arg is None:
-            await self.push(syntaxerr)
-            return
+            return await self.push(syntaxerr)
         arg = self._strip_command_keyword('TO:', arg)
         if arg is None:
-            await self.push(syntaxerr)
-            return
+            return await self.push(syntaxerr)
         address, params = self._getaddr(arg)
+        if address is None:
+            return await self.push("553 5.1.3 Error: malformed address")
         if not address:
-            await self.push(syntaxerr)
-            return
+            return await self.push(syntaxerr)
         if not self.session.extended_smtp and params:
-            await self.push(syntaxerr)
-            return
+            return await self.push(syntaxerr)
         rcpt_options = params.upper().split()
         params = self._getparams(rcpt_options)
         if params is None:
-            await self.push(syntaxerr)
-            return
+            return await self.push(syntaxerr)
         # XXX currently there are no options we recognize.
         if len(params) > 0:
-            await self.push(
-                '555 RCPT TO parameters not recognized or not implemented')
-            return
+            return await self.push(
+                '555 RCPT TO parameters not recognized or not implemented'
+            )
+
         status = await self._call_handler_hook('RCPT', address, rcpt_options)
         if status is MISSING:
             self.envelope.rcpt_tos.append(address)
